#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#define SET_OUT 0

static int read_ls (const char *file, unsigned short **ls, int *order)
{        
	int len, retval, i, j, bg;
        FILE *f;

        f = fopen (file, "r");
        if (f == NULL) {
                printf ("Failed to open the file %s\n", file);
                return -1;
        }
        retval = 0;
        do {
                bg = fscanf (f, "%d\n", &len);
                if ((bg != 1) || (len <= 0)) {
                        printf ("Bad length value\n");
                        retval = -1;
                        break;
                }
                (*ls) = (unsigned short *) malloc (len * len * sizeof (unsigned short));
                if ((*ls) == NULL) {
                        printf ("Failed to allocate memory\n");
                        retval = -1;
                        break;
                }
                for (i = 0; i < len; i++) {
                        for (j = 0; j < len - 1; j++) {
                                fscanf (f, "%hu ", &((*ls) [len * i + j]));
                        }
                        fscanf (f, "%hu\n", &((*ls) [len * i + j]));
                }
        } while (0);

        (*order) = len;

        fclose (f);
        return retval;
}

static void print_ls (unsigned short *ls, int order)
{
        int i, j;

        for (i = 0; i < order; i++) {
                for (j = 0; j < order; j++) {
                        printf ("%hu ", ls [i * order + j]);
                }
                printf ("\n");
        }
}

static void find_sigma_inv (int order, unsigned short *ls, unsigned short *perm)
{
	int i;

	for (i = 0; i < order; i++) {
		perm [ls [i]] = i;
	}
	return;
}

static void fill_l1_row (int order, int rownum, unsigned short *ls, unsigned short *l1, unsigned short *perm)
{
	int i;

	for (i = 0; i < order; i++) {
		l1 [order * rownum + i] = ls [order * rownum + perm [i]];
	}
	return;
}

static int form_l1 (int order, unsigned short *ls, unsigned short **l1)
{
	int retval, i;
	unsigned short *perm;

	perm = NULL;
	retval = 0;

	do {
		// first allocate memory
		(*l1) = (unsigned short *) malloc (order * order * sizeof (unsigned short));
		if ((*l1) == NULL) {
			printf ("Failed to allocate memory\n");
			retval = -1;
			break;
		}
		perm = (unsigned short *) malloc (order * sizeof (unsigned short));
		if (perm == NULL) {
			printf ("Failed to allocate memory\n");
			retval = -1;
			break;
		}
		find_sigma_inv (order, ls, perm);
		
		// fill the matrix l1
		for (i = 0; i < order; i++) {
			fill_l1_row (order, i, ls, (*l1), perm);
		}
	} while (0);

	if (perm != NULL) {
		free (perm);
	}

	return retval;
}

static int form_l2 (int order, unsigned short **l1)
{
	unsigned short *l2;
	int i;

	l2 = (unsigned short *) malloc (order * order * sizeof (unsigned short));
	if (l2 == NULL) {
		printf ("Failed to allocate memory\n");
		return -1;
	}

	for (i = 0; i < order; i++) {
		memcpy (l2 + order * ((*l1) [order * i]), (*l1) + order * i, order * sizeof (short));
	}
	free (*l1);
	(*l1) = l2;
	return 0;
}

static int does_it_commute (int order, unsigned short *l1)
{
	int i, j, affine;

	affine = 1;

	for (i = 0; i < order - 1; i++) {
		for (j = i + 1; j < order; j++) {
			if (l1 [i * order + j] != l1 [j * order + i]) {
				affine = 0;
				break;
			}
		}
		if (affine == 0) {
			break;
		}
	}

	return affine;
}

static int is_it_associative (int order, unsigned short *l1)
{
	int affine, i, j, k;

	affine = 1;

	// a dumb associativity test - exhaust all triples
	for (i = 0; i < order; i++) {
		for (j = 0; j < order; j++) {
			for (k = 0; k < order; k++) {
				if (l1 [i * order + l1 [j * order + k]] !=
					l1 [l1 [i * order + j] * order + k]) {
					affine = 0;
					break;
				}
			}
			if (affine == 0) {
				break;
			}
		}
		if (affine == 0) {
			break;
		}
	}

	return affine;
}

static int get_alpha (int order, unsigned short *ls, 
				unsigned short zero, unsigned short **alpha)
{
	int i, j;

	(*alpha) = (unsigned short *) malloc (order * sizeof (unsigned short));
	if ((*alpha) == NULL) {
		printf ("Failed to allocate memory\n");
		return -1;
	}

	for (j = 0; j < order; j++) {
		if (ls [j] == zero) {
			break;
		}
	}
	if (j >= order) {
		return -1;
	}

	for (i = 0; i < order; i++) {
		(*alpha) [i] = ls [i * order + j];
	}

	return 0;
}

static int check_automorphism (int order, unsigned short *l1, unsigned short *alpha)
{
	int i, j, affine;

	affine = 1;

	// a dumb test - exhaust all possible pairs
	for (i = 0; i < order; i++) {
		for (j = 0; j < order; j++) {
			if (l1 [alpha [i] * order + alpha [j]] != 
						alpha [l1 [i * order + j]]) {
				affine = 0;
				break;
			}
		}
		if (affine == 0) {
			break;
		}
	}

	return affine;	
}

static int get_beta (int order, unsigned short *ls, 
				unsigned short zero, unsigned short **beta)
{
	int i, j;

	(*beta) = (unsigned short *) malloc (order * sizeof (unsigned short));
	if ((*beta) == NULL) {
		printf ("Failed to allocate memory\n");
		return -1;
	}

	for (i = 0; i < order; i++) {
		if (ls [i * order] == zero) {
			break;
		}
	}
	if (i >= order) {
		return -1;
	}

	for (j = 0; j < order; j++) {
		(*beta) [j] = ls [i * order + j];
	}

	return 0;
}

static int check_representation (int order, unsigned short *ls, 
				unsigned short *l1, unsigned short *alpha, 
							unsigned short *beta)
{
	unsigned short c;
	int affine, i, j;

	affine = 1;
	c = ls [0];

	// brute force check
	for (i = 0; i < order; i++) {
		for (j = 0; j < order; j++) {
			if (ls [i * order + j] != 
					l1 [l1 [alpha [i] * 
					order + beta [j]] * order + c]) {
				affine = 0;
				break;
			}
		}
		if (affine == 0) {
			break;
		}
	}

	return affine;
}

static int is_good_ls (int order, unsigned short *ls)
{
	unsigned short *board;
	int retval, i, j, sum;

	retval = 0;

	board = (unsigned short *) malloc (order * sizeof (unsigned short));
	if (board == NULL) {
		printf ("Failed to allocate memory\n");
		return -1;
	}

	for (i = 0; i < order; i++) {
		memset (board, 0, order * sizeof (unsigned short));
		for (j = 0; j < order; j++) {
			if (ls [i * order + j] >= order) {
				printf ("Element too large\n");
				retval = 1;
				break;
			}
			board [ls [i * order + j]] = 1;
		}
		if (retval == 1) {
			break;
		}
		sum = 0;
		for (j = 0; j < order; j++) {
			sum+= board [j];
		}
		if (sum < order) {
			printf ("Not a permutation\n");
			retval = 1;
			break;
		}
	}

	if (retval != 0) {
		free (board);
		return retval;
	}

	for (i = 0; i < order; i++) {
		memset (board, 0, order * sizeof (unsigned short));
		for (j = 0; j < order; j++) {
			board [ls [i * order + j]] = 1;
		}
		sum = 0;
		for (j = 0; j < order; j++) {
			sum+= board [j];
		}
		if (sum < order) {
			printf ("Not a permutation\n");
			retval = 1;
			break;
		}
	}

	free (board);
	return retval;
}

//int main (int argc, const char *argv [])
int TestAffine0(int order,int **r)
{
        unsigned short *ls, *l1, *alpha, *beta;
        int  retval, affine,i,j,k;

	affine = 0;
	l1 = NULL;
	alpha = NULL;
	beta = NULL;

ls=new unsigned short[order*(size_t)order];
for(k=0,i=0;i<order;i++)for(j=0;j<order;j++)ls[k++]=r[i][j];
	// walk through the original algorithm:
	do {
		retval = form_l1 (order, ls, &l1);
		if (retval < 0) {
			printf ("Failed to form l1\n");
			break;
		}
		// print_ls (l1, order);

		retval = form_l2 (order, &l1);
		if (retval < 0) {
			if(SET_OUT)printf ("Failed to form l2\n");
			break;
		}
		// print_ls (l1, order);
		affine = does_it_commute (order, l1);
		if (affine == 0) {
			if(SET_OUT)printf ("Commutativity test failed\n");
			break;
		}
		affine = is_it_associative (order, l1);
		if (affine == 0) {
			if(SET_OUT)printf ("Associativity test failed\n");
			break;
		}
		retval = get_alpha (order, ls, l1 [0], &alpha);
		if (retval < 0) {
			if(SET_OUT)printf ("Failed to get alpha\n");
			break;
		}
		affine = check_automorphism (order, l1, alpha);
		if (affine == 0) {
			if(SET_OUT)printf ("Alpha is not an automorphism\n");
			break;
		}
		retval = get_beta (order, ls, l1 [0], &beta);
		if (retval < 0) {
			if(SET_OUT)printf ("Failed to get beta\n");
			break;
		}
		affine = check_automorphism (order, l1, beta);
		if (affine == 0) {
			if(SET_OUT)printf ("Beta is not an automorphism\n");
			break;
		}
		affine = check_representation (order, ls, l1, alpha, beta);
  if(!affine)
  {if(SET_OUT)printf ("check_representation==0\n");}
	} while (0);

	if (l1 != NULL) {
		free (l1);
	}
	if (alpha != NULL) {
		free (alpha);
	}
	if (beta != NULL) {
		free (beta);
	}
	//free (ls);

	/*if (retval != 0) {
		printf ("System error, result not obtained\n");
	} else {
		if (affine == 0) {
			printf ("The quasigroup is not affine\n");
		} else {
			printf ("The quasigroup is affine\n");
		}
	}//*/
 delete[] ls;
	return affine;
}
