#include "fun.h"
#include "simple.h"
extern int **r__;
extern int SetOutputSubclasses;

void wpair_list_init (wpair_list_t *l)
{
	(*l) = NULL;
}

int wpair_list_add (wpair_list_t *l, unsigned int first, unsigned int second)
{
	wpair_t *wpair;

	wpair = (wpair_t *) malloc (sizeof (wpair_t));
	if (wpair == NULL) {
		printf ("Failed to allocate memory\n");
		return -1;
	}
	wpair->first = first;
	wpair->second = second;
	wpair->prev = NULL;
	if ((*l) == NULL) {
		wpair->next = NULL;
		(*l) = wpair;
	} else {
		wpair->next = *l;
		(*l)->prev = wpair;
		(*l) = wpair;
	}
	return 0;
}

void wpair_list_remove (wpair_list_t *l, wpair_t *wpair)
{
	if ((wpair->next == NULL) && (wpair->prev == NULL)) {
		(*l) = NULL;
	} else {
		if (wpair->next != NULL) {
			(wpair->next)->prev = wpair->prev;
		}
		if (wpair->prev != NULL) {
			(wpair->prev)->next = wpair->next;
		}
		if (wpair->prev == NULL) {
			(*l) = wpair->next;
		}
	}
}

void wpair_list_deinit (wpair_list_t *l)
{
	wpair_t *wpair;

	while ((*l) != NULL) {
		wpair = *l;
		wpair_list_remove (l, wpair);
		free (wpair);
	}
}

void eqclass_list_init (eqclass_list_t *l)
{
	l->head = NULL;
	l->tail = NULL; 
}

int eqclass_list_add (eqclass_list_t *l, unsigned int num)
{
	eqclass_t *item;

	item = (eqclass_t *) malloc (sizeof (eqclass_t));
	if (item == NULL) {
		printf ("Failed to allocate memory\n");
                return -1;
        }

	item->rep = num;
	if (l->head == NULL) {
		item->next = NULL;
		l->head = item;
		l->tail = item;
	} else {
		item->next = l->head;
		l->head = item;
	}
	return 0;
}

void eqclass_list_remove (eqclass_list_t *l, eqclass_t *eqclass)
{
	if (eqclass->next == NULL) {
		l->head = NULL;
		l->tail = NULL;
	} else {
		l->head = eqclass->next;
	}
}

void eqclass_list_deinit (eqclass_list_t *l, int do_free)
{
	eqclass_t *item;

	while (l->head != NULL) {
		item = l->head;
		eqclass_list_remove (l, item);
		if (do_free != 0) {
			free (item);
		}
	}
}

// add l2 to the end of l
void eqclass_list_merge (eqclass_list_t *l, eqclass_list_t *l2)
{
	l->tail->next = l2->head;
	l->tail = l2->tail;
	eqclass_list_deinit (l2, 0);
}

int read_ls (const char *file, unsigned int **ls, int *order)
{
	int len, retval, i, j, bg;
	FILE *f;

	f = fopen (file, "r");
	if (f == NULL) {
		printf ("Failed to open the file %s\n", file);
		return -1;
	}
	retval = 0;
	do {
		bg = fscanf (f, "%d\n", &len);
		if ((bg != 1) || (len <= 0)) {
			printf ("Bad length value\n");
			retval = -1;
			break;
		}
		(*ls) = (unsigned int *) malloc (len * len * sizeof (unsigned int));
		if ((*ls) == NULL) {
			printf ("Failed to allocate memory\n");
			retval = -1;
			break;
		}
		for (i = 0; i < len; i++) {
			for (j = 0; j < len - 1; j++) {
				fscanf (f, "%u ", &((*ls) [len * i + j]));
			}
			fscanf (f, "%u\n", &((*ls) [len * i + j]));
		}
	} while (0);

	(*order) = len;

	fclose (f);
	return retval;
}

void print_ls (unsigned int *ls, int order)
{
	int i, j;

	for (i = 0; i < order; i++) {
		for (j = 0; j < order; j++) {
			printf ("%u ", ls [i * order + j]);
		}
		printf ("\n");
	}
}

// 1 is simple, 0 is nonsimple, -1 is error
int is_simple (unsigned int *ls, int order)
{
	int retval, i, k, maxsize, c1, c2, *sizes;
	wpair_list_t l;
	wpair_t *cur;
	unsigned int LSeed, *Class, v1, v2;
	char *equiv;
	eqclass_list_t *classlist;
	eqclass_t *eq1, *eq2;

	Class = (unsigned int *) malloc (order * sizeof (unsigned int));
	if (Class == NULL) {
		printf ("Failed to allocate memory\n");
		return -1;
	}
	equiv = NULL;
	sizes = NULL;
	classlist = NULL;

	do {
		equiv = (char *) malloc (order * order);
		if (equiv == NULL) {
			printf ("Failed to allocate memory\n");
			retval = -1;
			break;
		}
		sizes = (int *) malloc (order * sizeof (int));
		if (sizes == NULL) {
			printf ("Failed to allocate memory\n");
			retval = -1;
			break;
		}
		classlist = (eqclass_list_t *) malloc (order * sizeof (eqclass_list_t));
		if (classlist == NULL) {
			printf ("Failed to allocate memory\n");
			retval = -1;
			break;
		}

		for (LSeed = 1; LSeed < order; LSeed++) {
			// main cycle - exhaust seeds
			// first initialize data
			wpair_list_init (&l);
			retval = wpair_list_add (&l, 0, LSeed);
			for (i = 0; i < order; i++) {
				Class [i] = i;
				sizes [i] = 1;
				eqclass_list_init (&(classlist [i]));
				if (i != LSeed) {
					eqclass_list_add (&(classlist [i]), i);
				}
			}
			eqclass_list_add (&(classlist [0]), LSeed);
			Class [LSeed] = 0;
			sizes [0] = 2;
			sizes [LSeed] = 0;
			memset (equiv, 0, order * order);
			equiv [LSeed] = 1;
			equiv [LSeed * order] = 1;
			maxsize = 2;
			retval = 0;
			// then construct the transitive closure
			while ((l != NULL) && (maxsize <= order / 2)) {
				cur = l;
				wpair_list_remove (&l, cur);
				for (i = 0; i < order; i++) {
					v1 = ls [i * order + cur->first];
					v2 = ls [i * order + cur->second];
					// is this wpair known to be equivalent?
					if (equiv [v1 * order + v2] == 0) {
						if (retval != 0) {
							break;
						}
						// merge classes
						c1 = Class [v1];
						c2 = Class [v2];
						if (sizes [c1] + sizes [c2] > maxsize) {
							maxsize = sizes [c1] + sizes [c2];
						}
						for (eq1 = classlist [c1].head; eq1 != NULL; eq1 = eq1->next) {
                            Class [eq1->rep] = c2;
							for (eq2 = classlist [c2].head; eq2 != NULL; eq2 = eq2->next) {
								equiv [eq1->rep * order + eq2->rep] = 1;
								equiv [eq1->rep + order * eq2->rep] = 1;
								retval = wpair_list_add (&l, eq1->rep, eq2->rep);
								if (retval != 0) {
									printf ("Failed to add a wpair\n");
									break;
								}
							}
						}
						eqclass_list_merge (&(classlist [c2]), &classlist [c1]);

						sizes [c2] += sizes [c1];
						sizes [c1] = 0;
					}
					v1 = ls [i + cur->first * order];
					v2 = ls [i + cur->second * order];
					// is this wpair known to be equivalent?
					if (equiv [v1 * order + v2] == 0) {
						if (retval != 0) {
							break;
						}
						// merge classes
						c1 = Class [v1];
						c2 = Class [v2];
						if (sizes [c1] + sizes [c2] > maxsize) {
							maxsize = sizes [c1] + sizes [c2];
						}

						for (eq1 = classlist [c1].head; eq1 != NULL; eq1 = eq1->next) {
                            Class [eq1->rep] = c2;
							for (eq2 = classlist [c2].head; eq2 != NULL; eq2 = eq2->next) {
								equiv [eq1->rep * order + eq2->rep] = 1;
								equiv [eq1->rep + order * eq2->rep] = 1;
								retval = wpair_list_add (&l, eq1->rep, eq2->rep);
								if (retval != 0) {
									printf ("Failed to add a wpair\n");
									break;
								}
							}
						}
						eqclass_list_merge (&(classlist [c2]), &classlist [c1]);

						sizes [c2] += sizes [c1];
						sizes [c1] = 0;
					}
				}
				free (cur);
				if (retval < 0) {
					break;
				}
			}
			if (l != NULL) {
				wpair_list_deinit (&l);
			}

			if ((retval >= 0) && (maxsize > order / 2)) {
				// printf ("Check %d: simple\n", LSeed);
				retval = 1;
			} else {
				// printf ("Check %d: nonsimple\n", LSeed);
				if(SetOutputSubclasses){O1 fprintf (f,"Congruence obtained:\n"); O2}
				k = 1;
				if(SetOutputSubclasses)
				{O1
				for (i = 0; i < order; i++) {
					if (classlist [i].head != NULL) {
						fprintf (f,"Class number %d:\n", k);
						k++;
						for (eq1  = classlist [i].head; eq1 != NULL; eq1 = eq1->next) {
							fprintf (f," %u", eq1->rep);
						}
						fprintf (f,"\n");
						//eqclass_list_deinit (&(classlist [i]), 1);
					}
				}O2
				}					//----W
  if(0){//printf("AG>\n"); //if(1)for(int idom=0;idom<nSub[0];idom++){printf("[%d]:",idom);for(i=1;i<=n;i++)if(nSub[i]==idom)printf("%d ",i-1); printf("\n");}
   int **gr=(int**)allocate_2D_Array(order,order+8, sizeof(int),1); int ierr,ngr=0,i;
//   for(idom=0,ngr=0;idom<nSub[0];idom++,ngr++){for(i=1;i<=n;i++)if(nSub[i]==idom)gr[ngr][++gr[ngr][0]]=i-1;}
				for (ngr=0, i = 0; i < order; i++) {
					if (classlist [i].head != NULL) {
						//printf ("Class number %d:\n", k);
						ngr++;
						for (eq1  = classlist [i].head; eq1 != NULL; eq1 = eq1->next) {
							gr[ngr-1][++gr[ngr-1][0]]=eq1->rep;//printf (" %u", eq1->rep);
						}
					}
				}
   if((ierr=CheckNotPrime(r__,gr,ngr,order))!=0){printf("Error %d: AG: not correct domains\n",ierr);printM("r:",r__,order);exit(-1);}
   delete[] (char*)gr;
  }
				for (i = 0; i < order; i++) {if (classlist [i].head != NULL) {eqclass_list_deinit (&(classlist [i]), 1);}}
					//----W
				break;
			}
			for (i = 0; i < order; i++) {
				if (classlist [i].head != NULL) {
					eqclass_list_deinit (&(classlist [i]), 1);
				}
			}

		} 
	} while (0);
	free (Class);
	if (equiv != NULL) {
		free (equiv);
	}
	if (sizes != NULL) {
		free (sizes);
	}
	if (classlist != NULL) {
		free (classlist);
	}

	return retval;
}

